__all__ = [
    '__project__',
    '__title__',
    '__author__',
    '__copyright__',
    '__license__',
    '__version__',
    '__mail__',
    '__maintainer__',
    '__status__']

__project__ = 'PyDMD'
__title__ = "pydmd"
__author__ = "Nicola Demo, Marco Tezzele, Francesco Andreuzzi"
__copyright__ = "Copyright 2017-2021, PyDMD contributors"
__license__ = "MIT"
__version__ = "0.4.0.post2206"
__mail__ = 'demo.nicola@gmail.com, marcotez@gmail.com, andreuzzi.francesco@gmail.com'
__maintainer__ = __author__
__status__ = "Stable"
